@@ -1,13 +1,20 @@
-{-# LANGUAGE CPP, DeriveDataTypeable, GeneralizedNewtypeDeriving, FlexibleContexts #-}
+{-# LANGUAGE CPP,
+  DeriveDataTypeable, GeneralizedNewtypeDeriving,
+  DeriveFunctor, DeriveFoldable, DeriveTraversable,
+  FlexibleContexts #-}
 
 {-| Abstract names should carry unique identifiers and stuff. Not right now though.
 -}
 module Agda.Syntax.Abstract.Name where
 
 import Control.Monad.State
-import Data.Generics (Typeable, Data)
+
+import Data.Foldable (Foldable)
+import Data.Traversable (Traversable)
+import Data.Typeable (Typeable)
 import Data.List
 import Data.Function
+import Data.Hashable
 
 import Agda.Syntax.Position
 import Agda.Syntax.Common
@@ -31,7 +38,7 @@ data Name = Name { nameId	   :: NameId
 		 , nameBindingSite :: Range
 		 , nameFixity	   :: Fixity'
 		 }
-    deriving (Typeable, Data)
+    deriving (Typeable)
 
 -- | Qualified names are non-empty lists of names. Equality on qualified names
 --   are just equality on the last name, i.e. the module part is just
@@ -42,32 +49,44 @@ data Name = Name { nameId	   :: NameId
 data QName = QName { qnameModule :: ModuleName
 		   , qnameName	 :: Name
 		   }
-    deriving (Typeable, Data)
+    deriving (Typeable)
+
+-- | Something preceeded by a qualified name.
+data QNamed a = QNamed
+  { qname  :: QName
+  , qnamed :: a
+  }
+  deriving (Typeable, Show, Functor, Foldable, Traversable)
 
 -- | A module name is just a qualified name.
 --
 -- The 'SetRange' instance for module names sets all individual ranges
 -- to the given one.
 newtype ModuleName = MName { mnameToList :: [Name] }
-  deriving (Eq, Ord, Typeable, Data)
+  deriving (Eq, Ord, Typeable)
 
 -- | Ambiguous qualified names. Used for overloaded constructors.
 --
 -- Invariant: All the names in the list must have the same concrete,
 -- unqualified name.
 newtype AmbiguousQName = AmbQ { unAmbQ :: [QName] }
-  deriving (Typeable, Data, HasRange, Show)
+  deriving (Typeable, HasRange, Show)
 
 instance HasRange ModuleName where
   getRange (MName []) = noRange
   getRange (MName xs) = getRange xs
 
+-- | A module is anonymous if the qualification path ends in an underscore.
+isAnonymousModuleName :: ModuleName -> Bool
+isAnonymousModuleName (MName []) = False
+isAnonymousModuleName (MName ms) = C.isNoName $ nameConcrete $ last ms
+
 -- | Sets the ranges of the individual names in the module name to
 -- match those of the corresponding concrete names. If the concrete
 -- names are fewer than the number of module name name parts, then the
 -- initial name parts get the range 'noRange'.
 --
--- @C.D.E `withRangesOf` [A, B]@ returns @C.D.E@ but with ranges set
+-- @C.D.E \`withRangesOf\` [A, B]@ returns @C.D.E@ but with ranges set
 -- as follows:
 --
 -- * @C@: 'noRange'.
@@ -193,7 +212,7 @@ freshNoName r =
 freshNoName_ :: (MonadState s m, HasFresh NameId s) => m Name
 freshNoName_ = freshNoName noRange
 
--- | Get the next version of the concrete name. For instance, @nextName "x" = "x'"@.
+-- | Get the next version of the concrete name. For instance, @nextName "x" = "x₁"@.
 --   The name must not be a 'NoName'.
 nextName :: Name -> Name
 nextName x = x { nameConcrete = C.Name noRange $ nextSuf ps }
@@ -219,6 +238,10 @@ instance Ord Name where
 instance Show Name where
   show x = show (nameConcrete x) -- ++ "|" ++ show (nameId x)
 
+instance Hashable Name where
+  {-# INLINE hash #-}
+  hash = hash . nameId
+
 instance Show QName where
   show q = concat $ intersperse "." $ map show $ qnameToList q
 
@@ -231,6 +254,10 @@ instance Eq QName where
 instance Ord QName where
   compare = compare `on` qnameName
 
+instance Hashable QName where
+  {-# INLINE hash #-}
+  hash = hash . qnameName
+
 instance HasRange Name where
   getRange = getRange . nameConcrete
 
