@@ -9,6 +9,7 @@
 module Agda.Termination.CallGraph
   ( -- * Structural orderings
     Order(Mat), decr
+  , increase, decrease
   , (.*.)
   , supremum, infimum
   , decreasing, le, lt, unknown, orderMat
@@ -89,6 +90,17 @@ instance Show Order where
 instance HasZero Order where
   zeroElement = Unknown
 
+-- | Raw increase which does not cut off.
+increase :: Int -> Order -> Order
+increase i o = case o of
+  Unknown -> Unknown
+  Decr k  -> Decr $ k - i
+  Mat m   -> Mat $ fmap (increase i) m
+
+-- | Raw decrease which does not cut off.
+decrease :: Int -> Order -> Order
+decrease i o = increase (-i) o
+
 -- | Smart constructor for @Decr k :: Order@ which cuts off too big values.
 --
 -- Possible values for @k@: @- ?cutoff '<=' k '<=' ?cutoff + 1@.
@@ -214,10 +226,14 @@ okM :: Matrix Integer Order -> Matrix Integer Order -> Bool
 okM m1 m2 = (rows $ size m2) == (cols $ size m1)
 
 -- | The supremum of a (possibly empty) list of 'Order's.
-
+--   More information (i.e., more decrease) is bigger.
+--   'Unknown' is no information, thus, smallest.
 supremum :: (?cutoff :: Int) => [Order] -> Order
 supremum = foldr maxO Unknown
 
+-- | @('Order', 'maxO', '.*.')@ forms a semiring, with 'Unknown' as
+-- zero and 'Le' as one.
+
 maxO :: (?cutoff :: Int) => Order -> Order -> Order
 maxO o1 o2 = case (o1,o2) of
                (Decr k, Decr l) -> Decr (max k l) -- cut off not needed
@@ -227,14 +243,11 @@ maxO o1 o2 = case (o1,o2) of
                (Mat m,_) -> maxO (collapse m) o2
                (_,Mat m) ->  maxO o1 (collapse m)
 
--- | @('Order', 'max', '.*.')@ forms a semiring, with 'Unknown' as
--- zero and 'Le' as one.
-
 -- | The infimum of a (non empty) list of 'Order's.
-
+--  'Unknown' is the least element, thus, dominant.
 infimum :: (?cutoff :: Int) => [Order] -> Order
-infimum (o:l) = foldl minO o l
-infimum [] = __IMPOSSIBLE__
+infimum (o:l) = foldl' minO o l
+infimum []    = __IMPOSSIBLE__
 
 minO :: (?cutoff :: Int) => Order -> Order -> Order
 minO o1 o2 = case (o1,o2) of
@@ -436,6 +449,12 @@ union :: Monoid meta
       => CallGraph meta -> CallGraph meta -> CallGraph meta
 union cs1 cs2 = CallGraph $ (Map.unionWith mappend `on` cg) cs1 cs2
 
+-- | 'CallGraph' is a monoid under 'union'.
+
+instance Monoid meta => Monoid (CallGraph meta) where
+  mempty  = empty
+  mappend = union
+
 -- | Inserts a call into a call graph.
 
 insert :: Monoid meta
