@@ -1,4 +1,4 @@
-{-# LANGUAGE CPP #-}
+{-# LANGUAGE CPP, DeriveFunctor #-}
 
 {- | Sparse matrices.
 
@@ -108,7 +108,7 @@ prop_Arbitrary_MIx = mIxInvariant
 -- | Type of matrices, parameterised on the type of values.
 
 data Matrix i b = M { size :: Size i, unM :: [(MIx i, b)] }
-  deriving (Eq, Ord)
+  deriving (Eq, Ord, Functor)
 
 matrixInvariant :: (Num i, Ix i) => Matrix i b -> Bool
 matrixInvariant m = all (\ (MIx i j, b) -> 1 <= i && i <= rows sz
@@ -334,7 +334,7 @@ mul :: (Enum i, Num i, Ix i, Eq a)
     => Semiring a -> Matrix i a -> Matrix i a -> Matrix i a
 mul semiring m1 m2 = M (Size { rows = rows (size m1), cols = cols (size m2) }) $
   filter (\ (i,b) -> b /= Semiring.zero semiring) $
-  [ (MIx i j, foldl (Semiring.add semiring) (Semiring.zero semiring) $
+  [ (MIx i j, List.foldl' (Semiring.add semiring) (Semiring.zero semiring) $
                 map snd $ interAssocWith (Semiring.mul semiring) v w)
     | (i,v) <- toSparseRows m1
     , (j,w) <- toSparseRows $ transpose m2 ]
