@@ -29,14 +29,15 @@ type Stack = [(CompiledClauses, MaybeReducedArgs, Args -> Args)]
 match :: CompiledClauses -> MaybeReducedArgs -> (Args -> Args) -> Stack -> TCM (Reduced (Blocked Args) Term)
 match Fail args patch stack = return $ NoReduction $ NotBlocked (patch $ map ignoreReduced args)
 match (Done xs t) args _ _
-  | m < n     = return $ YesReduction $ substs (reverse $ toTm args)
+  | m < n     = return $ YesReduction $ applySubst (parallelS $ reverse $ toTm args)
                                       $ foldr lam t (drop m xs)
-  | otherwise = return $ YesReduction $ substs (reverse $ toTm args0) t `apply` map ignoreReduced args1
+  | otherwise = return $ YesReduction $
+                  applySubst (parallelS $ reverse $ toTm args0) t `apply` map ignoreReduced args1
   where
     n              = length xs
     m              = length args
     toTm           = map (unArg . ignoreReduced)
-    (args0, args1) = splitAt n args
+    (args0, args1) = splitAt n $ map (fmap $ fmap shared) args
     lam x t        = Lam (argHiding x) (Abs (unArg x) t)
 match (Case n bs) args patch stack =
   case genericSplitAt n args of
@@ -56,22 +57,24 @@ match (Case n bs) args patch stack =
           patchCon c m args = patch (args0 ++ [Arg h r $ Con c vs] ++ args1)
             where (args0, args1') = splitAt n args
                   (vs, args1)     = splitAt m args1'
-      case w of
+      case ignoreSharing <$> w of
         Blocked x _            -> return $ NoReduction $ Blocked x (patch $ map ignoreReduced args')
         NotBlocked (MetaV x _) -> return $ NoReduction $ Blocked x (patch $ map ignoreReduced args')
         NotBlocked (Lit l) -> case Map.lookup l (litBranches bs) of
           Nothing -> match' stack''
           Just cc -> match cc (args0 ++ args1) patchLit stack''
           where
-            stack'' = (++ stack') $ case cv of
+            stack'' = (++ stack') $ case ignoreSharing cv of
               Con c vs -> case Map.lookup c (conBranches bs) of
                 Nothing -> []
-                Just cc -> [(cc, args0 ++ map (MaybeRed red) vs ++ args1, patchCon c (length vs))]
+                Just cc -> [(content cc, args0 ++ map (MaybeRed red) vs ++ args1, patchCon c (length vs))]
               _        -> []
         NotBlocked (Con c vs) -> case Map.lookup c (conBranches bs) of
           Nothing -> match' stack'
-          Just cc -> match cc (args0 ++ map (MaybeRed red) vs ++ args1)
-                              (patchCon c (length vs)) stack'
+          Just cc -> match (content cc)
+                           (args0 ++ map (MaybeRed red) vs ++ args1)
+                           (patchCon c (length vs))
+                           stack'
         NotBlocked _ -> return $ NoReduction $ NotBlocked (patch $ map ignoreReduced args')
 
 match' :: Stack -> TCM (Reduced (Blocked Args) Term)
@@ -79,7 +82,6 @@ match' ((c, args, patch):stack) = match c args patch stack
 match' [] = typeError $ GenericError "Incomplete pattern matching"
 
 unfoldCorecursion v = case v of
-  -- Andreas, 2011-10-03 removing DontCare here allow matching on irrelevant things
-  DontCare v -> unfoldCorecursion v
   Def f args -> unfoldDefinition True unfoldCorecursion (Def f []) f args
+  Shared{}   -> fmap shared <$> unfoldCorecursion (ignoreSharing v) -- don't update when unfolding corecursion!
   _          -> reduceB v
