@@ -25,6 +25,7 @@ import Agda.TypeChecking.MetaVars
 import {-# SOURCE #-} Agda.TypeChecking.Conversion
 import Agda.TypeChecking.Pretty
 import Agda.TypeChecking.Constraints
+import Agda.TypeChecking.Polarity
 import Agda.Utils.List
 import Agda.Utils.Monad
 import Agda.Utils.Permutation
@@ -38,20 +39,27 @@ reduceHead v = ignoreAbstractMode $ do
   -- first, possibly rewrite literal v to constructor form
   v <- constructorForm v
   reportSDoc "tc.inj.reduce" 30 $ text "reduceHead" <+> prettyTCM v
-  case v of
+  case ignoreSharing v of
     Def f args -> do
+      let v0 = Def f []
       def <- theDef <$> getConstInfo f
       case def of
---         Function{ funClauses = [ _ ] }  -> unfoldDefinition False reduceHead v f args
-        Datatype{ dataClause = Just _ } -> unfoldDefinition False reduceHead v f args
-        Record{ recClause = Just _ }    -> unfoldDefinition False reduceHead v f args
+        -- Andreas, 2012-11-06 unfold aliases (single clause terminating functions)
+        -- see test/succeed/Issue747
+        -- We restrict this to terminating functions to not make the
+        -- type checker loop here on non-terminating functions.
+        -- see test/fail/TerminationInfiniteRecord
+        Function{ funClauses = [ _ ], funDelayed = NotDelayed, funTerminates = Just True }
+                                        -> unfoldDefinition False reduceHead v0 f args
+        Datatype{ dataClause = Just _ } -> unfoldDefinition False reduceHead v0 f args
+        Record{ recClause = Just _ }    -> unfoldDefinition False reduceHead v0 f args
         _                               -> return $ notBlocked v
     _ -> return $ notBlocked v
 
 headSymbol :: Term -> TCM (Maybe TermHead)
 headSymbol v = ignoreAbstractMode $ do
   v <- ignoreBlocking <$> reduceHead v
-  case v of
+  case ignoreSharing v of
     Def f _ -> do
       def <- theDef <$> getConstInfo f
       case def of
@@ -61,17 +69,24 @@ headSymbol v = ignoreAbstractMode $ do
           -- Don't treat axioms in the current mutual block
           -- as constructors (they might have definitions we
           -- don't know about yet).
-          fs <- lookupMutualBlock =<< currentMutualBlock
+          fs <- lookupMutualBlock =<< currentOrFreshMutualBlock
           if Set.member f fs
             then return Nothing
             else return (Just $ ConHead f)
-        _           -> return Nothing
+        Function{}    -> return Nothing
+        Primitive{}   -> return Nothing
+        Constructor{} -> __IMPOSSIBLE__
     Con c _ -> return (Just $ ConHead c)
     Sort _  -> return (Just SortHead)
     Pi _ _  -> return (Just PiHead)
     Lit _   -> return Nothing -- handle literal heads as well? can't think of
                               -- any examples where it would be useful...
-    _       -> return Nothing
+    Lam{}   -> return Nothing
+    Var{}   -> return Nothing
+    Level{} -> return Nothing
+    MetaV{} -> return Nothing
+    DontCare{} -> return Nothing
+    Shared{}   -> __IMPOSSIBLE__
 
 checkInjectivity :: QName -> [Clause] -> TCM FunctionInverse
 checkInjectivity f cs
@@ -115,7 +130,7 @@ checkInjectivity f cs = do
 
 -- | Argument should be on weak head normal form.
 functionInverse :: Term -> TCM InvView
-functionInverse v = case v of
+functionInverse v = case ignoreSharing v of
   Def f args -> do
     d <- theDef <$> getConstInfo f
     case d of
@@ -130,6 +145,10 @@ data InvView = Inv QName Args (Map TermHead Clause)
 
 useInjectivity :: Comparison -> Type -> Term -> Term -> TCM ()
 useInjectivity cmp a u v = do
+  reportSDoc "tc.inj.use" 30 $ fsep $
+    pwords "useInjectivity on" ++
+    [ prettyTCM u, prettyTCM cmp, prettyTCM v, text ":", prettyTCM a
+    ]
   uinv <- functionInverse u
   vinv <- functionInverse v
   case (uinv, vinv) of
@@ -183,8 +202,7 @@ useInjectivity cmp a u v = do
             ]
           -- and this is the order the variables occur in the patterns
           let ms' = permute (invertP $ compactP perm) ms
-          cxt <- getContextTelescope
-          let sub = (reverse ms ++ idSub cxt)
+          let sub = parallelS (reverse ms)
           margs <- runReaderT (evalStateT (metaArgs ps) ms') sub
           reportSDoc "tc.inj.invert" 20 $ vcat
             [ text "inversion"
@@ -194,7 +212,10 @@ useInjectivity cmp a u v = do
               , text "type =" <+> prettyTCM ftype
               ]
             ]
-          pol <- getPolarity' cmp f
+          -- Since we do not care for the value of non-variant metas here,
+          -- we can treat 'Nonvariant' as 'Invariant'.
+          -- That ensures these metas do not remain unsolved.
+          pol <- purgeNonvariant <$> getPolarity' cmp f
           -- The clause might not give as many patterns as there
           -- are arguments (point-free style definitions).
           let args' = take (length margs) args
@@ -219,7 +240,7 @@ useInjectivity cmp a u v = do
                , text "which does not expose a constructor"
                ]
              patternViolation
-        `catchError` \err -> case errError err of
+        `catchError` \err -> case err of
           TypeError   {} -> throwError err
           Exception   {} -> throwError err
           IOException {} -> throwError err
@@ -234,7 +255,7 @@ useInjectivity cmp a u v = do
     dotP :: Monad m => Term -> StateT [Term] (ReaderT Substitution m) Term
     dotP v = do
       sub <- ask
-      return $ substs sub v
+      return $ applySubst sub v
 
     metaArgs args = mapM metaArg args
     metaArg arg = traverse metaPat arg
